/* ------------------------------------------------------------------------
 *
 *        File: smgsortf.c
 *     Project: Source Mapper.
 *     Created: June 8, 1992.
 * Description: Modul for  sortere elementer i fil.
 *
 * Copyright (C) 2000 Leif-Erik Larsen.
 * This file is part of the Source Mapper source package.
 * Source Mapper is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, in version 2 as it comes in the
 * "COPYING" file of the XWorkplace main distribution.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * ------------------------------------------------------------------------ */




#include "smg.h"




static int  construct        ( FILE *stream, uint16 size );
static int  destruct         ( FILE *stream );
static int  get1element_file ( FILE *stream, uint16 nr, char *buff );
static int  put1element_file ( FILE *stream, uint16 nr, char *buff );
static void get1element_buff ( uint16 nr, char *buff );
static void put1element_buff ( uint16 nr, char *buff );
static int  sort_file        ( FILE *stream );
static void sort_buff        ( void );
static int  sort_file_buff   ( FILE *stream );




static char  *mainbuff;        /* Hovedbuffer for lagring av fildata */
static char  *buff1, *buff2;   /* Buffer for mellomlagring ved sammenlikning */
static long   buffsize;        /* Antall byte som er allokert til buffer */
static uint16 tot_elements;    /* Totalt antall elementer i filen */
static uint16 element_size;    /* Antall byte i hvert element */
static uint16 buff_elements;   /* Antall elementer som er lagret i bufferet */
static int    all_data_in_buff;/* TRUE hvis hele filen er lagret i bufferet */




static int construct ( FILE *stream, uint16 size )
/* Opprettet: Fredag 16. oktober 1992.
   Parameter: "stream" Filen som skal sorteres.
              "size"   Antall byte i hvert element i filen.
   Retur    : E/O.
   Beskriv  : Initierer globale variable til denne modulen. I initieringen
              inngr ogs tildeling av hukommelse fra far hop til buffer,
              samt  stille filpeker til stream til start av filen.
*/
{
   unsigned long memfree;              /* Ledig RAM */
   long          flen;                 /* Antall byte i fil */

   buff1 = NULL;
   buff2 = NULL;
   mainbuff = NULL;

   if (!Rewind (stream))               /* Sett filpeker til start av filen */
   {
      return (ERROR);
   }

   element_size = size;                /* Antall byte i hvert element */

   buff1 = Alloc (size + 1);           /* Buffer1 for mellomlager */
   buff2 = Alloc (size + 1);           /* Buffer2 for mellomlager */

   if (buff1 == NULL || buff2 == NULL)
   {
      if (buff1) Free (buff1);
      if (buff2) Free (buff2);
      RETERR (14, NULL);               /* Out of memory! */
   }

   memfree = MemLeft ();               /* Ledig RAM */

   if (memfree < 4000L)                /* Hvis for lite ledig RAM */
   {
      mainbuff = NULL;
      buffsize = 0;
      buff_elements = 0;
      all_data_in_buff = FALSE;

      return (OK); /* Ingen feil, men sortering m gjres p disk */
   }

   if (memfree > 65535L * size)
      memfree = 65535L * size;

   buffsize = memfree - 1000;          /* Antall byte som er allokert */
   mainbuff = Alloc (buffsize);        /* Alloker hukommelse fra heap */
   flen     = FLength (stream);        /* Antall byte i filen */

   if (flen == -1L)                    /* Hvis feil ved lesing av fillengde */
   {
      return (ERROR);
   }

   tot_elements = (uint16) (flen / size); /* Antall elementer i fil */

   /* Hvis det er plass til samtlige filelementer i buffer: */
   if (tot_elements * size <= buffsize)
   {
      buff_elements    = tot_elements;
      all_data_in_buff = TRUE;
   }

   /* Hvis det ikke er plass til samtlige filelementer i buffer: */
   else
   {
      buff_elements    = (uint16) (buffsize / size);
      all_data_in_buff = FALSE;
   }

   return (OK);
} /* construct (); */




static int destruct ( FILE *stream )
/* Opprettet: Fredag 16. oktober 1992.
   Parameter: "stream" Filen hvor filpeker skal stilles til filstart.
   Retur    : E/O.
   Beskriv  : Frigjr allokert hukommelse, og stiller filpeker til
              oppgitt fil til filstart.
*/
{
   if (mainbuff) Free (mainbuff);
   if (buff2)    Free (buff2);
   if (buff1)    Free (buff1);

   mainbuff = NULL;
   buff1    = NULL;
   buff2    = NULL;

   if (!Rewind (stream))
   {
      return (ERROR);
   }

   return (OK);
} /* destruct (); */




static int get1element_file ( FILE *stream, uint16 nr, char *buff )
/* Opprettet: Tirsdag 2. juni 1992.
   Parameter: "stream" Filen som element skal leses fra.
              "nr"     Element nummer (1..).
              "buff"   Buffer hvor elementet leses inn.
   Retur    : E/O.
   Beskriv  : Brukes nr sorteringen gjres p disk.
              Leser inn et oppgitt element fra fil.
              Filpeker forlates uten  stilles tilbake til det sted den
              stod ved start.
*/
{
   if (!FSeek (stream, (nr - 1) * element_size, SEEK_SET))
   {
      return (ERROR);
   }

   if (FRead (buff, element_size, 1, stream) != 1)
   {
      return (ERROR);
   }

   return (OK);
} /* get1element_file (); */




static int put1element_file ( FILE *stream, uint16 nr, char *buff )
/* Opprettet: Tirsdag 2. juni 1992.
   Parameter: "stream" Filen som element skal lagres til.
              "nr"     Element nummer (1..).
              "buff"   Buffer som lagrer aktuelt element.
   Retur    : E/O.
   Beskriv  : Brukes nr sorteringen gjres p disk.
              Skriver et oppgitt element til fil.
              Filpeker forlates uten  stilles tilbake til det sted den
              stod ved start.
*/
{
   if (!FSeek (stream, (nr - 1) * element_size, SEEK_SET))
   {
      return (ERROR);
   }

   if (FWrite (buff, element_size, 1, stream) != 1)
   {
      return (ERROR);
   }

   return (OK);
} /* put1element_file (); */




static void get1element_buff ( uint16 nr, char *buff )
/* Opprettet: Fredag 16. oktober 1992.
   Parameter: "nr"   Element nummer (1..).
              "buff" Buffer hvor oppgitt elementet kopieres til.
   Retur    :
   Beskriv  : Brukes nr sorteringen gjres p buffer.
              Returnerer oppgitt element fra buffer.
*/
{
   memcpy (buff, ADRESS (mainbuff, nr - 1, element_size), element_size);
} /* get1element_buff (); */




static void put1element_buff ( uint16 nr, char *buff )
/* Opprettet: Fredag 16. oktober 1992.
   Parameter: "nr"   Element nummer (1..).
              "buff" Buffer som skal kopieres til hovedbuffer.
   Retur    :
   Beskriv  : Brukes nr sorteringen gjres p buffer.
              Kopierer oppgitt element til hovedbuffer.
*/
{
   memcpy (ADRESS (mainbuff, nr - 1, element_size), buff, element_size);
} /* put1element_buff (); */




static int sort_file ( FILE *stream )
/* Opprettet: Fredag 16. oktober 1992.
   Parameter: "stream" Filen som skal sorteres.
   Retur    : E/O.
   Beskriv  : Sorterer elementene i filen.
              Sorteringen utfres direkte p filen, uten bruk av andre
              buffere enn mellomlagring av innleste elementer.
              Denne metoden brukes dersom det ikke er nok ledig RAM til
               allokere noe buffer i det hele tatt.
*/
{
   int  cmpres;                        /* Resultat av elementsammenlikning */
   long gap, i, j, k;                  /* For shell-sorteringen */

   gap = tot_elements / 2;
   while (gap > 0)
   {
      for (i = gap + 1; i <= tot_elements; i++)
      {
         j = i - gap;

         while (j > 0)
         {
            k = j + gap;

            if (!get1element_file (stream, (uint16) j, buff1) ||
                !get1element_file (stream, (uint16) k, buff2))
            {
                return (ERROR);
            }

            cmpres = memcmp (buff1, buff2, element_size);

            if (cmpres <= 0)           /* Hvis buff1 <= buff2 */
            {
               j = 0;
            }

            else                       /* Hvis buff1 > buff2 */
            {
               if (!put1element_file (stream, (uint16) j, buff2) ||
                   !put1element_file (stream, (uint16) k, buff1))
               {
                   return (ERROR);
               }

               j -= gap;
            }
         }
      }

      gap /= 2;
   }

   return (OK);
} /* sort_file (); */




static void sort_buff ( void )
/* Opprettet: Fredag 16. oktober 1992.
   Parameter:
   Retur    :
   Beskriv  : Sorterer elementene i "mainbuff".
              Denne metoden brukes dersom samtlige filelementer fr
              plass i bufferet, eller for  sortere innholdet i bufferet nr
              bare deler av filelementene fr plass.
              Merk: Filelementene m vre lest in til "mainbuff" p forrhnd.
                    Det sorterte buffer m dessuten skrives til filen
                    utenfor denne funksjonen.
*/
{
   int  cmpres;                        /* Resultat av elementsammenlikning */
   long gap, i, j, k;                  /* For shell-sorteringen */

   gap = buff_elements / 2;
   while (gap > 0)
   {
      for (i = gap + 1; i <= buff_elements; i++)
      {
         j = i - gap;

         while (j > 0)
         {
            k = j + gap;

            get1element_buff ((uint16) j, buff1);
            get1element_buff ((uint16) k, buff2);

            cmpres = memcmp (buff1, buff2, element_size);

            if (cmpres <= 0)           /* Hvis buff1 <= buff2 */
            {
               j = 0;
            }

            else                       /* Hvis buff1 > buff2 */
            {
               put1element_buff ((uint16) j, buff2);
               put1element_buff ((uint16) k, buff1);

               j -= gap;
            }
         }
      }

      gap /= 2;
   }
} /* sort_buff (); */




static int sort_file_buff ( FILE *stream )
/* Opprettet: Fredag 16. oktober 1992.
   Parameter:
   Retur    : E/O.
   Beskriv  : Sorterer elementene i filen.
              Sorteringen utfres p flgende mte:
              Frst:    Blokker av filen sorteres i buffer.
              Deretter: Hele filen sorteres direkte p disk.
                        N er allerede det meste av filen sortert, og
                        sorteringen p disk gr raskere.
              Denne metoden brukes dersom det allokerte buffer bare er
              stort nok for deler av filementene om gangen.
*/
{
   long   lastfpos;                    /* For  huske gammel filposisjon */
   uint16 max_buffelements = buff_elements;
   uint16 elements_left = tot_elements;/* Antall gjennvrende ellementer */
   uint16 elements_read;               /* Antall elementer som skal leses */

   if (!Rewind (stream))
   {
      return (ERROR);
   }

   while (elements_left)               /* S lenge det er elementer igjen */
   {
      /* Finn hvor mange elementer som skal leses i ved neste lesing: */
      if (elements_left > max_buffelements)
      {
         /* Fyll opp buffer: */
         elements_read = max_buffelements;
      }
      else
      {
         /* Resten av elementene: */
         elements_read = elements_left;
      }

      /* Husk filposisjon, og fyll opp buffer: */
      if ((lastfpos = FTell (stream)) == -1L)
      {
         return (ERROR);
      }

      if (FRead (mainbuff, element_size,
                           elements_read, stream) != elements_read)
      {
         return (ERROR);
      }

      buff_elements = elements_read;   /* Antall elementer i bufferet */
      sort_buff ();                    /* Sorter innholdet i bufferet */

      /* Plasser filpeker tilbake til der buffer ble lest fra: */
      if (!FSeek (stream, lastfpos, SEEK_SET))
      {
         return (ERROR);
      }

      /* Skriv sortert blokk til fil: */
      if (FWrite (mainbuff, element_size,
                            elements_read, stream) != elements_read)
      {
         return (ERROR);
      }

      elements_left -= elements_read;
   }

   return (OK);
} /* sort_file_buff (); */




int sortfile ( FILE *stream, uint16 size )
/* Opprettet: Tirsdag 2. juni 1992.
   Parameter: "stream"  Filen som skal sorteres.
              "size"    Antall byte til hvert element i filen.
   Retur    : E/O.
   Beskriv  : Sorterer elementene i oppgitt fil. Filen forutsettes pnet,
              for bde skriving og lesing i binrt mode.
              Filpeker til "stream" plasseres helt i starten av filen,
              etter vellykket sortering i denne funksjon. Sorteringen
              utfres med den kjente shell-allgoritmen.
              Merk: Funksjonen kan ikke sortere fler enn 65535 elementer!
*/
{
   /* Initier globale variable og alloker minne til buffere fra hop,
      samt still filpeker til stream til filstart: */
   if (!construct (stream, size))
   {
      destruct (stream);
      return (ERROR);
   }

   if (buffsize == 0)                  /* Hvis ingen buffer er allokert */
   {
      /* Sorter filen direkte p disk: */
      if (!sort_file (stream))
      {
         goto END_ERR;
      }
   }

   else
   if (all_data_in_buff)               /* Hele filen fr plass i buffer */
   {
      /* Sorter filen direkte i allokert buffer: */
      if (!Rewind (stream) ||
          FRead  (mainbuff, element_size, tot_elements, stream) != tot_elements)
      {
         goto END_ERR;
      }

      sort_buff ();

      if (!Rewind (stream) ||
          FWrite (mainbuff, element_size, tot_elements, stream) != tot_elements)
      {
         goto END_ERR;
      }
   }

   else                                /* Bare deler av filen fr plass */
   {
      /* Sorter filen delvis i buffer og delvis p disk: */
      if (!sort_file_buff (stream))
      {
         goto END_ERR;
      }
   }

   /* Frigjr allokert hukommelse, og plasser filpeker til start av fil: */
   if (!destruct (stream))
   {
      return (ERROR);
   }

   return (OK);

END_ERR:

   destruct (stream);

   RETURN_ERR;
} /* sortfile (); */

